var _ = require('lodash');
var async = require('async');

module.exports = (function() {

  function Registry() {
    this.lookupTable = {};
    this.register({
      name: 'describe',
      method: (cb) => {
        cb(null, _.values(this.lookupTable));
      },
      description: 'Returns all currently available methods'
    });
  }

  Registry.prototype.freeze = function() {
    _.forEach(this.lookupTable, function(value) {
       Object.freeze(value);
    });
    this.lookupTable = Object.freeze(this.lookupTable);
    Object.freeze(this);
  };

  Registry.prototype.lookup = function(methodName) {
    if(!_.has(this.lookupTable, methodName)) return false;

    return _.clone(this.lookupTable[methodName]);
  };

  Registry.prototype._register = function(methodName, method, description) {
    this.lookupTable[methodName] = {
      run: method,
      // Set length here so that future partial param application will not
      // obfuscate the expected number of params
      length: method.length - 1, // Don't count callback
      name: methodName,
      description: description || ''
    };
  };

  Registry.prototype._registerOne = function(methodDef) {
    var methodName = methodDef.name;
    var methods = methodDef.method || methodDef.methods;

    if(!_.isString(methodName)) {
      throw new Error('Method name provided for registration must be a string');
    }

    if(_.isArray(methods)) {
      if(methods.length === 0) {
        throw new Error('Method array must contain at least one method to be registered');
      }

      if(methods.length === 1) {
        return this._register(methodName, methods[0]);
      }

      var asyncMethods = _.map(methods, function(method) {
        if(!_.isFunction(method)) {
          throw new Error('Method must be a function');
        }
        return async.ensureAsync(method);
      });

      var firstMethodLength = methods[0].length;
      this._register(methodName, function() {
        var cb = _.last(arguments);
        var tmpMethods = _.clone(asyncMethods);
        tmpMethods[0] = _.bind(tmpMethods[0], this,
          // TODO: determine if this leaks argumnents
          _.slice(arguments, 0, arguments.length-1)
        );
        tmpMethods = _.map(tmpMethods, (method) => _.bind(method, this));
        async.waterfall(tmpMethods, cb);
      });

      // Reset param length to be that of first method
      this.lookupTable[methodName].length = firstMethodLength;

    } else if(_.isFunction(methods)) {
      this._register(methodName, methods, methodDef.description);
    } else {
      throw new Error('Method must be a function or an array of functions');
    }
  };

  Registry.prototype._registerMany = function(methodDefs) {
    if(!_.isArray(methodDefs)) {
      throw new Error('Object must be an array of method definitions');
    }

    _.forEach(methodDefs, (methodDef) => {
      this._registerOne(methodDef.methodName, methodDef.method || methodDef.methods);
    });
  };

  Registry.prototype.register = function(one, two) {
    if(_.isString(one) && _.isObject(two)) {
      this._registerOne({
        name: one,
        method: two
      });
    } else if(_.isObject(one) && two == null) {
      if(_.isArray(one)) {
        this._registerMany(one);
      } else {
        this._registerOne(one);
      }
    } else {
      throw new Error('Invalid parameters supplied to register');
    }
  };

  return Registry;
})();
